#include "e-filesel.h"

#include <config.h>

#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <gconf/gconf-client.h>
#include <gtk/gtk.h>
#include <string.h>

struct _EFileSelPrivate {
  GtkTreeStore * store;
  GtkTreeModel * model;
  GtkWidget * treeview;

  gchar * uri;

  GtkWidget * rbutton;
  GtkWidget * newbutton;
  GtkWidget * ohmenu;
  GtkWidget * netbutton;

  GtkWidget * fentry;
  GtkWidget * foomenu;

  GtkWidget * fmenu;
  GtkWidget * filter_list_menu;

  GtkTooltips * tooltips;

  GConfClient * client;
  gboolean save_as_dialog;
  gboolean firstfile;
  gchar * filter_type;
};

enum {
  PROP_ZERO,
  PROP_FILENAME
};

enum {
  FILTER_ALL,
  FILTER_AUDIO,
  FILTER_IMAGE,
  FILTER_TEXT,
  FILTER_VIDEO
};

enum {
  RESPONSE,
  LAST_SIGNAL
};

static GtkWindowClass * parent_class = NULL;
static gboolean SaveAsDialog = FALSE;
static guint signals [LAST_SIGNAL] = { 0, };

static void e_filesel_class_init (EFileSelClass * klass);
static void e_filesel_init (EFileSel * filesel);
static void e_filesel_finalize (GObject * object);
static void e_filesel_destroy (GtkObject * object);
static void e_filesel_map (GtkWidget * widget);

static gint e_filesel_filecmp (GtkTreeModel * model,
			       GtkTreeIter * a,
			       GtkTreeIter * b,
			       EFileSel * filesel);
static void e_filesel_load_files (GnomeVFSAsyncHandle * handle,
				  GnomeVFSResult result,
				  GList * list,
				  guint entries_read,
				  EFileSel * filesel);
static void e_filesel_recent_goto (GtkMenuShell * shell,
				   EFileSel * filesel);
static void e_filesel_add_recent (EFileSel * filesel,
				  const gchar * uri);
static void e_filesel_go_to_subdir (GtkTreeView * treeview,
				    GtkTreePath * path,
				    GtkTreeViewColumn * column,
				    EFileSel * filesel);
static void e_filesel_entry_activate (GtkEntry * entry,
				      EFileSel * filesel);
static void e_filesel_refresh_dir (GtkWidget * widget,
				   EFileSel * filesel);
static void e_filesel_file_selected (GtkTreeSelection * selection,
				     EFileSel * filesel);
static void e_filesel_filter (GtkMenuShell * shell,
			      EFileSel * filesel);
static void e_filesel_init_filters (EFileSel * filesel);



GType e_filesel_get_type (void) {
  static GType filesel_type = 0;

  if (!filesel_type) {
    GTypeInfo filesel_info = {
      sizeof (EFileSelClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) e_filesel_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,
      sizeof (EFileSel),
      0,
      (GInstanceInitFunc) e_filesel_init,
      NULL
    };

    filesel_type = g_type_register_static (GTK_TYPE_WINDOW,
					   "EFileSel",
					   &filesel_info, 0);
  }
  return filesel_type;
}

static void e_filesel_class_init (EFileSelClass * klass) {
  GtkWindowClass * object_class = GTK_WINDOW_CLASS (klass);

  parent_class = g_type_class_peek (GTK_TYPE_WINDOW);

  signals[RESPONSE] = g_signal_new ("response",
				    G_TYPE_FROM_CLASS (klass),
				    G_SIGNAL_RUN_LAST,
				    G_STRUCT_OFFSET (EFileSelClass, response),
				    NULL, NULL,
				    g_cclosure_marshal_VOID__INT,
				    G_TYPE_NONE, 1, G_TYPE_INT);

  klass->response = NULL;
}

static void e_filesel_cancel_clicked (GtkButton * button,
				      EFileSel * filesel) {
  g_signal_emit (filesel, signals[RESPONSE], 0, GTK_RESPONSE_CANCEL);
  gtk_widget_destroy (GTK_WIDGET (filesel));
}

static void e_filesel_ok_clicked (GtkButton * button,
				  EFileSel * filesel) {
  g_signal_emit (filesel, signals[RESPONSE], 0, GTK_RESPONSE_OK);
}

static gint e_filesel_delete_event (GtkWidget * widget,
				    GdkEventAny * event,
				    gpointer userdata) {
  g_signal_emit (widget, signals[RESPONSE], 0, GTK_RESPONSE_CANCEL);
  return FALSE;
}

static void e_filesel_init (EFileSel * filesel) {
  GtkWidget * hbox, * vbox, * scrolley;
  GtkWidget * dbox, * label, * button;
  GtkCellRenderer * renderer;
  GtkTreeViewColumn * column;
  GdkPixbuf * pixbuf;

  filesel->priv = g_new0 (EFileSelPrivate, 1);

  filesel->priv->firstfile = TRUE;
  filesel->priv->filter_type = NULL;
  filesel->priv->save_as_dialog = SaveAsDialog;

  filesel->priv->client = gconf_client_get_default ();

  g_signal_connect (G_OBJECT (filesel), "delete_event",
		    G_CALLBACK (e_filesel_delete_event), NULL);

  dbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (filesel), dbox);
  
  gtk_window_set_type_hint (GTK_WINDOW (filesel), GDK_WINDOW_TYPE_HINT_DIALOG);
  gtk_window_set_position (GTK_WINDOW (filesel), GTK_WIN_POS_CENTER_ON_PARENT);

  if (filesel->priv->save_as_dialog) {
    gtk_window_set_title (GTK_WINDOW (filesel), "Save As");
  } else {
    gtk_window_set_title (GTK_WINDOW (filesel), "Open");
  }
  gtk_widget_set_usize (GTK_WIDGET (filesel), 300, 324);

  if (filesel->priv->save_as_dialog) {
    pixbuf = gtk_widget_render_icon (GTK_WIDGET (filesel), GTK_STOCK_SAVE_AS,
				     GTK_ICON_SIZE_MENU, NULL);
  } else {
    pixbuf = gtk_widget_render_icon (GTK_WIDGET (filesel), GTK_STOCK_OPEN,
				     GTK_ICON_SIZE_MENU, NULL);
  }
  gtk_window_set_icon (GTK_WINDOW (filesel), pixbuf);

  hbox = gtk_hbutton_box_new ();
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbox), GTK_BUTTONBOX_END);
  gtk_box_pack_end (GTK_BOX (dbox), hbox, FALSE, FALSE, 0);
  
  button = gtk_button_new_from_stock (GTK_STOCK_HELP);
  gtk_container_set_border_width (GTK_CONTAINER (button), 4);
  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  gtk_widget_show (button);
  gtk_button_box_set_child_secondary (GTK_BUTTON_BOX (hbox), button, TRUE);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

  button = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
  gtk_container_set_border_width (GTK_CONTAINER (button), 4);
  gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  gtk_widget_show (button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (e_filesel_cancel_clicked), filesel);

  if (filesel->priv->save_as_dialog) {
    button = gtk_button_new_from_stock (GTK_STOCK_SAVE);
  } else {
    button = gtk_button_new_from_stock (GTK_STOCK_OPEN);
  }
  gtk_container_set_border_width (GTK_CONTAINER (button), 4);
  gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  gtk_widget_show (button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (button);
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (e_filesel_ok_clicked), filesel);

  gtk_widget_show (hbox);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (dbox), hbox,
		      FALSE, FALSE, 0);

  filesel->priv->rbutton = gtk_button_new ();
  gtk_container_set_border_width (GTK_CONTAINER (filesel->priv->rbutton), 4);
  gtk_box_pack_start (GTK_BOX (hbox), filesel->priv->rbutton, FALSE, FALSE, 0);
  label = gtk_image_new_from_stock (GTK_STOCK_REFRESH, GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER (filesel->priv->rbutton), label);
  gtk_widget_show (label);
  gtk_widget_show (filesel->priv->rbutton);

  g_signal_connect (G_OBJECT (filesel->priv->rbutton), "clicked",
		    G_CALLBACK (e_filesel_refresh_dir), filesel);

  filesel->priv->ohmenu = gtk_option_menu_new ();
  gtk_container_set_border_width (GTK_CONTAINER (filesel->priv->ohmenu), 4);
  gtk_container_add (GTK_CONTAINER (hbox), filesel->priv->ohmenu);
  gtk_widget_show (filesel->priv->ohmenu);

  filesel->priv->foomenu = gtk_menu_new ();
  gtk_option_menu_set_menu (GTK_OPTION_MENU (filesel->priv->ohmenu),
			    filesel->priv->foomenu);
  g_signal_connect (G_OBJECT (filesel->priv->foomenu), "deactivate",
		    G_CALLBACK (e_filesel_recent_goto), filesel);

  filesel->priv->netbutton = gtk_button_new ();
  gtk_container_set_border_width (GTK_CONTAINER (filesel->priv->netbutton), 4);
  gtk_box_pack_end (GTK_BOX (hbox), filesel->priv->netbutton, FALSE, FALSE, 0);
  label = gtk_image_new_from_stock (GTK_STOCK_JUMP_TO, GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER (filesel->priv->netbutton), label);
  gtk_widget_show (label);
  gtk_widget_show (filesel->priv->netbutton);
  /*
    This button needs to pop up a menu of network places or something
  */
  gtk_widget_set_sensitive (filesel->priv->netbutton, FALSE);
  
  gtk_widget_show (hbox);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (dbox), hbox, TRUE, TRUE, 0);

  scrolley = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolley),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolley),
				       GTK_SHADOW_IN);
  gtk_box_pack_start (GTK_BOX (hbox), scrolley, TRUE, TRUE, 0);
  gtk_widget_show (scrolley);

  filesel->priv->treeview = gtk_tree_view_new ();
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (filesel->priv->treeview),
				     FALSE);
  filesel->priv->model = GTK_TREE_MODEL (gtk_tree_store_new (4, G_TYPE_STRING,
							     GDK_TYPE_PIXBUF,
							     G_TYPE_STRING,
							     G_TYPE_STRING));
  gtk_tree_view_set_model (GTK_TREE_VIEW (filesel->priv->treeview),
			   filesel->priv->model);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (filesel->priv->treeview), TRUE);

  g_signal_connect (G_OBJECT (filesel->priv->treeview), "row-activated",
		    G_CALLBACK (e_filesel_go_to_subdir), filesel);

  g_signal_connect (G_OBJECT (gtk_tree_view_get_selection (GTK_TREE_VIEW (filesel->priv->treeview))), "changed",
		    G_CALLBACK (e_filesel_file_selected), filesel);

  column = gtk_tree_view_column_new ();
  renderer = gtk_cell_renderer_pixbuf_new ();
  gtk_tree_view_column_pack_start (column, renderer, FALSE);
  gtk_tree_view_column_set_attributes (column, renderer,
				       "pixbuf", 1,
				       NULL);
		
  renderer = gtk_cell_renderer_text_new ();
  gtk_tree_view_column_pack_start (column, renderer, FALSE);
  gtk_tree_view_column_set_attributes (column, renderer,
				       "text", 2,
				       NULL);
  gtk_tree_view_column_set_title (column, "File Name");

  gtk_tree_view_append_column (GTK_TREE_VIEW (filesel->priv->treeview),
			       column);

  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Mime Type",
						     renderer,
						     "text", 3,
						     NULL);

  gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (filesel->priv->model), 2,
				   (GtkTreeIterCompareFunc) e_filesel_filecmp,
				   filesel, NULL);
  gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (filesel->priv->model),
					3, GTK_SORT_DESCENDING);
  gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (filesel->priv->model),
					2, GTK_SORT_ASCENDING);

  filesel->priv->store = GTK_TREE_STORE (filesel->priv->model);

  gtk_container_add (GTK_CONTAINER (scrolley), filesel->priv->treeview);
  gtk_widget_show (filesel->priv->treeview);

  gtk_widget_show (hbox);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 4);
  gtk_box_pack_start (GTK_BOX (dbox), hbox, FALSE, FALSE, 0);

  vbox = gtk_vbox_new (TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, TRUE, 0);

  label = gtk_label_new ("File _Types:");
  gtk_label_set_use_underline (GTK_LABEL (label), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_container_add (GTK_CONTAINER (vbox), label);
  gtk_widget_show (label);

  label = gtk_label_new ("File _Name:");
  gtk_label_set_use_underline (GTK_LABEL (label), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_container_add (GTK_CONTAINER (vbox), label);
  gtk_widget_show (label);

  gtk_widget_show (vbox);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 0);

  filesel->priv->fmenu = gtk_option_menu_new ();
  gtk_container_set_border_width (GTK_CONTAINER (filesel->priv->fmenu), 2);
  gtk_box_pack_start (GTK_BOX (vbox), filesel->priv->fmenu, TRUE, TRUE, 0);
  gtk_widget_show (filesel->priv->fmenu);

  e_filesel_init_filters (filesel);

  label = gtk_hbox_new (TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (label), 2);
  gtk_container_add (GTK_CONTAINER (vbox), label);

  filesel->priv->fentry = gtk_entry_new ();
  gtk_container_add (GTK_CONTAINER (label), filesel->priv->fentry);
  g_signal_connect (G_OBJECT (filesel->priv->fentry), "activate",
		    G_CALLBACK (e_filesel_entry_activate), filesel);
  gtk_widget_show (filesel->priv->fentry);
  gtk_widget_show (label);

  gtk_widget_show (vbox);
  gtk_widget_show (hbox);

  label = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX (dbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  gtk_widget_show (dbox);
}

GtkWidget * e_filesel_new (gboolean save_dialog) {

  SaveAsDialog = save_dialog;

  return GTK_WIDGET (g_object_new (E_TYPE_FILESEL, NULL));
}

void e_filesel_set_filename (EFileSel * filesel,
			     const gchar * filename) {
  gtk_entry_set_text (GTK_ENTRY (filesel->priv->fentry), filename);
}

void e_filesel_set_dirname (EFileSel * filesel,
			    const gchar * dirname) {
  GnomeVFSAsyncHandle * handle;

  if (filesel->priv->uri != NULL && *filesel->priv->uri) {
    g_free (filesel->priv->uri);
  }

  filesel->priv->uri = g_strdup (dirname);
  gtk_tree_store_clear (filesel->priv->store);
  gtk_tree_view_columns_autosize (GTK_TREE_VIEW (filesel->priv->treeview));
  gnome_vfs_async_load_directory (&handle, dirname,
				  GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
				  GNOME_VFS_FILE_INFO_FOLLOW_LINKS,
				  1, GNOME_VFS_PRIORITY_MAX,
				  (GnomeVFSAsyncDirectoryLoadCallback)
				  e_filesel_load_files, filesel);
  e_filesel_add_recent (filesel, dirname);
}

gchar * e_filesel_get_filename (EFileSel * filesel) {
  GtkTreeSelection * selection;
  GtkTreeIter iter;
  gchar * newuri, * mime_type, * returi;
  gchar * efname;

  efname = g_strdup (gtk_entry_get_text (GTK_ENTRY (filesel->priv->fentry)));
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (filesel->priv->treeview));

  if (gtk_tree_selection_get_selected (selection,
				       &filesel->priv->model,
				       &iter)) {
    gtk_tree_model_get (filesel->priv->model, &iter,
			0, &newuri,
			3, &mime_type, -1);
    if (!strncmp (mime_type, "x-directory/", strlen ("x-directory/"))) {
      if (newuri != NULL && *newuri && newuri[strlen (newuri) - 1] == '/') {
	returi = g_strconcat (newuri, efname, NULL);
      } else {
	returi = g_strconcat (newuri, "/", efname, NULL);
      }
    } else {
      if (!strcmp (g_basename (newuri), efname)) {
	returi = g_strdup (newuri);
      } else {
	if (filesel->priv->uri != NULL && *filesel->priv->uri &&
	    filesel->priv->uri[strlen (filesel->priv->uri) - 1] == '/') {
	  returi = g_strconcat (filesel->priv->uri, efname, NULL);
	} else {
	  returi = g_strconcat (filesel->priv->uri, "/", efname, NULL);
	}
      }
    }
  } else {
    if (filesel->priv->uri != NULL && *filesel->priv->uri &&
	filesel->priv->uri[strlen (filesel->priv->uri) - 1] == '/') {
      returi = g_strconcat (filesel->priv->uri, efname, NULL);
    } else {
      returi = g_strconcat (filesel->priv->uri, "/", efname, NULL);
    }
  }
  return returi;
}

void e_filesel_append_filter (EFileSel * filesel,
			      const gchar * filter_name,
			      const gchar * filter_type) {
  GtkWidget * menuitem;

  menuitem = gtk_menu_item_new_with_label (filter_name);
  gtk_object_set_data (GTK_OBJECT (menuitem), "mime_type",
		       g_strdup (filter_type));
  gtk_menu_append (GTK_MENU (filesel->priv->filter_list_menu), menuitem);
  gtk_widget_show (menuitem);
}

static void e_filesel_destroy (GtkObject * object) {
}

static void e_filesel_map (GtkWidget * widget) {
  GTK_WIDGET_CLASS (parent_class)->map (widget);
}

static void e_filesel_finalize (GObject * object) {
  EFileSel * filesel = E_FILESEL (object);

  g_free (filesel->priv->uri);
}

#define SHOW_HIDDEN_KEY "/apps/nautilus/preferences/show_hidden_files"
#define SHOW_BACKUP_KEY "/apps/nautilus/preferences/show_backup_files"
#define DIR_SORT_KEY "/apps/nautilus/preferences/sort_directories_first"
#define ICON_THEME_KEY "/desktop/gnome/file_views/icon_theme"

static gchar * e_filesel_nautilus_icon_crack (const gchar * icon,
					      const gchar * mime_type,
					      EFileSel * filesel) {
  gchar * newicon = NULL;
  gchar * themename = NULL;
  gchar * icontmp = NULL;
  gint i;

  themename = gconf_client_get_string (filesel->priv->client,
				       ICON_THEME_KEY, NULL);
  if (icon != NULL && *icon) {
    newicon = g_strconcat (NAUTIDIR, "/", themename, "/", icon, ".png", NULL);
    g_free (themename);
    if (g_file_test (newicon, G_FILE_TEST_EXISTS)) {
      return newicon;
    } else {
      g_free (newicon);
      newicon = g_strconcat (GPIXDIR, "/document-icons/", icon, ".png", NULL);
      if (g_file_test (newicon, G_FILE_TEST_EXISTS)) {
	return newicon;
      }
    }
    g_free (newicon);
    return NULL;
  } else if (!strncmp (mime_type, "x-directory/", strlen ("x-directory/"))) {
    if (themename != NULL && *themename) {
      newicon = g_strconcat (NAUTIDIR, "/", themename, "/i-directory.png",
			     NULL);
    } else {
      newicon = g_strconcat (GPIXDIR, "/gnome-folder.png", NULL);
    }
    g_free (themename);
    if (g_file_test (newicon, G_FILE_TEST_EXISTS)) {
      return newicon;
    }
    g_free (newicon);
    return NULL;
  } else {
    icontmp = g_strdup (mime_type);
    for (i = 0; i < strlen (icontmp); i++) {
      if (icontmp[i] == '/') {
	icontmp[i] = '-';
      }
    }
    if (themename != NULL && *themename) {
      newicon = g_strconcat (NAUTIDIR, "/", themename, "/gnome-",
			     icontmp, ".png", NULL);
      if (g_file_test (newicon, G_FILE_TEST_EXISTS)) {
	return newicon;
      } else {
	newicon = g_strconcat (GPIXDIR, "/document-icons/gnome-",
			       icontmp, ".png", NULL);
	if (g_file_test (newicon, G_FILE_TEST_EXISTS)) {
	  return newicon;
	}
      }
    }
    g_free (icontmp);
    g_free (themename);
    g_free (newicon);
    return NULL;
  }
}

static gint e_filesel_filecmp (GtkTreeModel * model,
			       GtkTreeIter * a,
			       GtkTreeIter * b,
			       EFileSel * filesel) {
  gchar * foo, * bar;
  gchar * foo_type, * bar_type;

  gtk_tree_model_get (filesel->priv->model, a, 2, &foo, 3, &foo_type, -1);
  gtk_tree_model_get (filesel->priv->model, b, 2, &bar, 3, &bar_type, -1);

  if (gconf_client_get_bool (filesel->priv->client, DIR_SORT_KEY, NULL)) {
    if (!strncmp (foo_type, "x-directory/", strlen ("x-directory/")) &&
	strncmp (bar_type, "x-directory/", strlen ("x-directory/"))) {
      return -1;
    } else if (strncmp (foo_type, "x-directory/", strlen ("x-directory/")) &&
	       !strncmp (bar_type, "x-directory/", strlen ("x-directory/"))) {
      return 1;
    } else {
      return strcmp (foo, bar);
    }
  } else {
    return strcmp (foo, bar);
  }
}


static void e_filesel_add_file (EFileSel * filesel,
				GnomeVFSFileInfo * fileinfo,
				gchar * newuri) {
  GtkTreeIter iter;
  const gchar * icon_file;
  gchar * newicon;
  GdkPixbuf * pixbuf = NULL, * icon = NULL;

  gtk_tree_store_append (filesel->priv->store, &iter, NULL);

  icon_file = (const gchar *) gnome_vfs_mime_get_icon (fileinfo->mime_type);
  if (icon_file != NULL && *icon_file) {
    if (!g_file_test (icon_file, G_FILE_TEST_EXISTS)) {
      newicon = e_filesel_nautilus_icon_crack (icon_file,
					       fileinfo->mime_type,
					       filesel);
      if (newicon != NULL && *newicon) {
	pixbuf = gdk_pixbuf_new_from_file (newicon, NULL);
	g_free (newicon);
      }
    } else {
      pixbuf = gdk_pixbuf_new_from_file (icon_file, NULL);
    }
  } else {
    newicon = e_filesel_nautilus_icon_crack (icon_file,
					     fileinfo->mime_type,
					     filesel);
    if (newicon != NULL && *newicon) {
      pixbuf = gdk_pixbuf_new_from_file (newicon, NULL);
      g_free (newicon);
    }
  }
  
  if (pixbuf != NULL) {
    icon = gdk_pixbuf_scale_simple (pixbuf, 16, 16, GDK_INTERP_HYPER);
    gtk_tree_store_set (filesel->priv->store, &iter,
			0, newuri,
			1, icon,
			2, fileinfo->name,
			3, fileinfo->mime_type,
			-1);
  } else {
    gtk_tree_store_set (filesel->priv->store, &iter,
			0, newuri,
			1, NULL,
			2, fileinfo->name,
			3, fileinfo->mime_type,
			-1);
  }
}

static void e_filesel_load_files (GnomeVFSAsyncHandle * handle,
				  GnomeVFSResult result,
				  GList * list,
				  guint entries_read,
				  EFileSel * filesel) {
  GnomeVFSFileInfo * fileinfo = NULL;
  gchar * newuri;

  if (result != GNOME_VFS_OK) {
    if (result != GNOME_VFS_ERROR_EOF)
      g_print ("Error: %s\n", gnome_vfs_result_to_string (result));
    return;
  } else {
    fileinfo = g_list_nth_data (list, 0);

    if (filesel->priv->firstfile) {
      gtk_tree_store_clear (filesel->priv->store);
      gtk_tree_view_columns_autosize (GTK_TREE_VIEW (filesel->priv->treeview));
      filesel->priv->firstfile = FALSE;
    }

    if (filesel->priv->uri[strlen (filesel->priv->uri) - 1] == '/') {
      newuri = g_strconcat (filesel->priv->uri, fileinfo->name, NULL);
    } else {
      newuri = g_strconcat (filesel->priv->uri, "/", fileinfo->name, NULL);
    }
    
    if (filesel->priv->filter_type != NULL && *filesel->priv->filter_type &&
	strncmp (fileinfo->mime_type, filesel->priv->filter_type,
		 strlen (filesel->priv->filter_type)) &&
	strncmp (fileinfo->mime_type, "x-directory/",
		 strlen ("x-directory/"))) {
      return;
    }

    if (strcmp (fileinfo->name, "..") && strcmp (fileinfo->name, ".")) {
      if (gconf_client_get_bool (filesel->priv->client,
				 SHOW_HIDDEN_KEY, NULL) &&
	  fileinfo->name[0] == '.') {
	if (gconf_client_get_bool (filesel->priv->client,
				   SHOW_BACKUP_KEY, NULL) &&
	    fileinfo->name[strlen (fileinfo->name) - 1] == '~') {
	  e_filesel_add_file (filesel, fileinfo, newuri);
	} else if (fileinfo->name[strlen (fileinfo->name) - 1] != '~') {
	  e_filesel_add_file (filesel, fileinfo, newuri);
	}
      } else if (fileinfo->name[0] != '.') {
	if (gconf_client_get_bool (filesel->priv->client,
				   SHOW_BACKUP_KEY, NULL) &&
	    fileinfo->name[strlen (fileinfo->name) - 1] == '~') {
	  e_filesel_add_file (filesel, fileinfo, newuri);
	} else if (fileinfo->name[strlen (fileinfo->name) - 1] != '~') {
	  e_filesel_add_file (filesel, fileinfo, newuri);
	}
      }
    }
  }

  while (gtk_events_pending ())
    gtk_main_iteration ();
}

static void e_filesel_recent_goto (GtkMenuShell * shell,
				   EFileSel * filesel) {
  GtkWidget * widget = NULL;
  GnomeVFSAsyncHandle * handle = NULL;
  GtkOptionMenu * omenu;
  gchar * uri = NULL;
  gint foo, i, bah;

  omenu = GTK_OPTION_MENU (filesel->priv->ohmenu);
  widget = gtk_menu_get_active (GTK_MENU (omenu->menu));

  if (widget != NULL) {
    foo = g_list_index (shell->children, widget);
  } else {
    foo = 0;
  }

  bah = g_list_length (shell->children);
  for (i = foo + 1; i < bah; bah--) {
    gtk_widget_destroy (g_list_nth_data (shell->children, i));
  }

  uri = g_strdup (gtk_object_get_data (GTK_OBJECT (widget), "uri"));
  if (strcmp (filesel->priv->uri, uri)) {
    if (filesel->priv->uri != NULL && *filesel->priv->uri) {
      g_free (filesel->priv->uri);
    }

    filesel->priv->uri = g_strdup (uri);
    filesel->priv->firstfile = TRUE;
    gnome_vfs_async_load_directory (&handle, filesel->priv->uri,
				    GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
				    GNOME_VFS_FILE_INFO_FOLLOW_LINKS,
				    1, GNOME_VFS_PRIORITY_MAX,
				    (GnomeVFSAsyncDirectoryLoadCallback)
				    e_filesel_load_files, filesel);
  }
  g_free (uri);
}

static void e_filesel_add_recent (EFileSel * filesel,
				  const gchar * uri) {
  const gchar * newuri;
  gchar * tmpuri = NULL;
  gchar * temp;
  GtkWidget * menuitem;
  GtkOptionMenu * omenu = GTK_OPTION_MENU (filesel->priv->ohmenu);
  gint foo;

  if (uri != NULL && *uri) {
    if (uri[strlen (uri) - 1] == '/') {
      newuri = tmpuri = g_strndup (uri, strlen (uri) - 1);
    } else {
      newuri = uri;
    }

    temp = strrchr (newuri, '/');
    if (temp) {
      newuri = temp + 1;
    }

    menuitem = gtk_menu_item_new_with_label (newuri);
    gtk_menu_append (GTK_MENU (filesel->priv->foomenu), menuitem);
    gtk_object_set_data (GTK_OBJECT (menuitem), "uri", filesel->priv->uri);
    gtk_widget_show (menuitem);

    foo = g_list_length (GTK_MENU (omenu->menu)->menu_shell.children);
    gtk_option_menu_set_history (omenu, foo - 1);

    g_free (tmpuri);
  }
}

static void e_filesel_go_to_subdir (GtkTreeView * treeview,
				    GtkTreePath * path,
				    GtkTreeViewColumn * column,
				    EFileSel * filesel) {
  GtkTreeIter iter;
  gchar * newuri, * mime_type;
  
  gtk_tree_model_get_iter (filesel->priv->model, &iter, path);
  gtk_tree_model_get (filesel->priv->model, &iter,
		      0, &newuri,
		      3, &mime_type, -1);

  if (!strncmp (mime_type, "x-directory/", strlen ("x-directory/"))) {
    e_filesel_set_dirname (filesel, newuri);
  } else {
    /* Open the fuggin file man */
  }
  g_free (newuri);
  g_free (mime_type);
}

static void e_filesel_entry_activate (GtkEntry * entry,
				      EFileSel * filesel) {
  gchar * uri;

  uri = g_strdup (gtk_entry_get_text (entry));

  if ((gchar *) gnome_vfs_get_mime_type (uri) != NULL) {
    if (!strncmp ((gchar *) gnome_vfs_get_mime_type (uri), "x-directory/",
		  strlen ("x-directory/"))) {
      e_filesel_set_dirname (filesel, uri);
    } else {
      /* Open the fuggin file */
    }
  }
  gtk_entry_set_text (GTK_ENTRY (filesel->priv->fentry), "");
  g_free (uri);
}

static void e_filesel_refresh_dir (GtkWidget * widget,
				   EFileSel * filesel) {
  GnomeVFSAsyncHandle * handle;

  filesel->priv->firstfile = TRUE;
  gnome_vfs_async_load_directory (&handle, filesel->priv->uri,
				  GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
				  GNOME_VFS_FILE_INFO_FOLLOW_LINKS,
				  1, GNOME_VFS_PRIORITY_MAX,
				  (GnomeVFSAsyncDirectoryLoadCallback)
				  e_filesel_load_files, filesel);
}

static void e_filesel_file_selected (GtkTreeSelection * selection,
				     EFileSel * filesel) {
  GtkTreeIter iter;
  gchar * foo, * bar;

  if (gtk_tree_selection_get_selected (selection, &filesel->priv->model,
				       &iter)) {
    gtk_tree_model_get (filesel->priv->model, &iter, 2, &foo, 3, &bar, -1);
    if (strncmp (bar, "x-directory/", strlen ("x-directory/"))) {
      gtk_entry_set_text (GTK_ENTRY (filesel->priv->fentry), foo);
    }
  } else {
    return;
  }
}

static void e_filesel_filter (GtkMenuShell * shell,
			      EFileSel * filesel) {
  GnomeVFSAsyncHandle * handle;
  GtkWidget * widget;
  GtkOptionMenu * omenu = GTK_OPTION_MENU (filesel->priv->fmenu);
  gint foo;

  widget = gtk_menu_get_active (GTK_MENU (omenu->menu));

  if (widget != NULL) {
    foo = g_list_index (shell->children, widget);
  } else {
    foo = 0;
  }
  switch (foo) {
  case FILTER_ALL:
    filesel->priv->filter_type = NULL;
    break;
  case FILTER_AUDIO:
    filesel->priv->filter_type = "audio/";
    break;
  case FILTER_IMAGE:
    filesel->priv->filter_type = "image/";
    break;
  case FILTER_TEXT:
    filesel->priv->filter_type = "text/";
    break;
  case FILTER_VIDEO:
    filesel->priv->filter_type = "video/";
    break;
  default:
    filesel->priv->filter_type = gtk_object_get_data (GTK_OBJECT (widget),
						      "mime_type");
    break;
  }
  filesel->priv->firstfile = TRUE;
  gnome_vfs_async_load_directory (&handle, filesel->priv->uri,
				  GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
				  GNOME_VFS_FILE_INFO_FOLLOW_LINKS,
				  1, GNOME_VFS_PRIORITY_MAX,
				  (GnomeVFSAsyncDirectoryLoadCallback)
				  e_filesel_load_files, filesel);
}

static void e_filesel_init_filters (EFileSel * filesel) {
  GtkWidget * menuitem;
  GtkOptionMenu * omenu = GTK_OPTION_MENU (filesel->priv->fmenu);

  filesel->priv->filter_list_menu = gtk_menu_new ();

  menuitem = gtk_menu_item_new_with_label ("All Files");
  gtk_menu_append (GTK_MENU (filesel->priv->filter_list_menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Audio Files");
  gtk_menu_append (GTK_MENU (filesel->priv->filter_list_menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Image Files");
  gtk_menu_append (GTK_MENU (filesel->priv->filter_list_menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Text Files");
  gtk_menu_append (GTK_MENU (filesel->priv->filter_list_menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Video Files");
  gtk_menu_append (GTK_MENU (filesel->priv->filter_list_menu), menuitem);
  gtk_widget_show (menuitem);
  
  gtk_option_menu_set_menu (omenu, filesel->priv->filter_list_menu);
  g_signal_connect (G_OBJECT (filesel->priv->filter_list_menu), "deactivate",
		    G_CALLBACK (e_filesel_filter), filesel);
}
